package net.wuenschenswert.intransparent;

import org.doit.muffin.*;
import org.doit.io.*;

import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.awt.*;

public class GifTransKillerFilter implements RequestFilter, ReplyFilter, ContentFilter
{
  Prefs prefs;
  GifTransKiller factory;
  InputObjectStream in = null;
  OutputObjectStream out = null;
  Request request;
  private InputStream gifInput;
  private int total;
  private ByteArray bytes;

  GifTransKillerFilter(GifTransKiller factory)
  {
    this.factory = factory;
  }

  public void setPrefs(Prefs prefs)
  {
    this.prefs = prefs;
  }

  public void filter(Request request) throws FilterException
  {
    this.request = request;
  }

  public void filter(Reply reply) throws FilterException
  {
// 	String s = reply.getContentType();
// 	if (s != null && s.startsWith("multipart/x-mixed-replace"))
// 	{
// 	    factory.process("Found server push " + request.getURL() + "\n");
// 	    throw new FilterException("Killed server push " + request.getURL());
// 	}
  }

  public boolean needsFiltration(Request request, Reply reply)
  {
    String s = reply.getContentType();
    if (s == null)
    {
      return false;
    }
    return s.startsWith("image/gif");
  }

  public void setInputObjectStream(InputObjectStream in)
  {
    this.in = in;
  }

  public void setOutputObjectStream(OutputObjectStream out)
  {
    this.out = out;
  }

  static final int MAX_BUF = 16384;

  public void run()
  {
    Thread.currentThread().setName("GifTransKiller");

    try
    {
      //assuming GIFs are small...  slurp it into memory.
      List/*<ByteArray>*/ objs = new ArrayList/*<ByteArray>*/();
      Object o;
      int total=0;
      while((o = in.read()) != null) {
        if(!(o instanceof ByteArray) || total > MAX_BUF) {
          // failed.
          // pipe through what we had buffered...
          for (Iterator it = objs.iterator(); it.hasNext();) {
            ByteArray red = (ByteArray) it.next();
            out.write(red);
          }
          // ... the offender...
          out.write(o);
          // ... and what else is coming.
          while((o = in.read()) != null)
            out.write(o);
          return;
        }
        ByteArray ba = (ByteArray) o;
        total += ba.length();
        objs.add(ba);
      }
      ByteArray bytes = new ByteArray(total);
      for (Iterator it = objs.iterator(); it.hasNext();) {
        ByteArray red = (ByteArray) it.next();
        bytes.append(red);
      }

      this.total = total;
      this.bytes = bytes;

      gifInput = new ByteArrayInputStream(bytes.bytes);

      try {
        parseGIF();
      } catch (Exception e) {
        e.printStackTrace();
      }

      out.write(bytes);

    } catch (IOException ioe) {
      ioe.printStackTrace();
    } finally {
      try
      {
        out.flush();
        out.close();
      } catch (IOException ioe) {
      }
    }
  }

  private void parseGIF() throws IOException {
    // <GIF Data Stream> ::= Header <Logical Screen> <Data>* Trailer

//    0x00-0x7F (0-127) are the Graphic Rendering blocks, excluding the Trailer (0x3B);
//    0x80-0xF9 (128-249) are the Control blocks;
//    0xFA-0xFF (250-255) are the Special Purpose blocks.

    if (!parseHeader())
      return;
    LSD lsd =  parseLogicalScreen();
    if (lsd == null)
      return;
    if (!parseDataSequence(lsd))
      return;
    if(read() != -1) {
      fail("expected eof");
      return;
    }
    // ok.
  }

  static final String magic = "GIF89a";

  private boolean parseHeader() throws IOException {
    for(int i=0; i<magic.length(); ++i)
      if(gifInput.read() != (int)magic.charAt(i)) {
        fail("no magic");
        return false;
      }
    return true;
  }

  private void fail(String s) {
    int pos = -1;
    try {
      if(gifInput != null)
        pos = currentPos();
    } catch (IOException e) {
    }
    request.addLogEntry("GifTransKiller", pos != -1 ? s+" at "+pos : s);
  }

  class LSD {
    int width;
    int height;
    int backgroundColorIndex;
    int pixelAspectRatio;
    int globalColorTableSize;  // 2 exponent, triplets
    int globalColorTablePos; // 0 if none
    boolean sorted;
    int colorResolutionBits;

    public void setGlobalColor(int index, byte[] rgb) {
      assert globalColorTablePos != 0;
      assert index >> globalColorTableSize == 0;
      int patchPos = globalColorTablePos + 3*index;
      int patchSize = 3;
      patch(patchPos, patchSize, rgb);
    }
  }

  private void patch(int patchPos, int patchSize, byte[] rgb) {
    System.arraycopy(rgb, 0, bytes.bytes, patchPos, patchSize);
  }

  private void patch(int patchPos, int b) {
    bytes.bytes[patchPos] = (byte) b;
  }

  private LSD parseLogicalScreen() throws IOException {
    // <Logical Screen> ::=      Logical Screen Descriptor [Global Color Table]
    LSD lsd = new LSD();
    lsd.width = readUnsigned();
    lsd.height = readUnsigned();
    int flags = read();
    lsd.backgroundColorIndex = read();
    lsd.pixelAspectRatio = read();

    if(lsd.pixelAspectRatio == -1) {
      fail("eof in logical screen descriptor");
      return null;
    }

    boolean hasGlobalColorTable = (flags & (1<<7)) != 0;
    lsd.colorResolutionBits = ((flags >> 4) & 7);
    lsd.sorted = (flags & (1<<3)) != 0; // most important first in global color table?
    lsd.globalColorTableSize = (flags & 7)+1;         // 2 exponent

    // Aspect Ratio = (Pixel Aspect Ratio + 15) / 64
    // ... if != 0

    if(hasGlobalColorTable) {
      lsd.globalColorTablePos = currentPos();
      int toSkip = 3 * (1 << lsd.globalColorTableSize);
      long skipped = gifInput.skip(toSkip);
      if(skipped != toSkip) {
        fail("eof in global color table");
        return null;
      }
    }
    return lsd;
  }

  private int readUnsigned() throws IOException {
    return read() + 256*read();
  }

  private int read() throws IOException {
    return gifInput.read();
  }

  private int currentPos() throws IOException {
    return total - gifInput.available();
  }

  private boolean parseDataSequence(LSD lsd) throws IOException {
    int b;
    while ((b = gifInput.read()) != TYPE_TRAILER) {
      Data data = parseData(b);
      if (data == null)
        return false;
      if(data instanceof TableBasedImage && ((TableBasedImage)data).graphicsControl != null) {
        GraphicsControl graphicsControl = ((TableBasedImage) data).graphicsControl;
        if(graphicsControl.transparentColorFlag && lsd.globalColorTablePos != 0) {
          Color color = prefs.getColor(GifTransKiller.PREF_COLOR);
          request.addLogEntry("GifTransKiller", "replaced transparent with "+color);
          byte[] rgb = { (byte) color.getRed(), (byte) color.getGreen(), (byte) color.getBlue() };
          lsd.setGlobalColor(graphicsControl.transparentColorIndex, rgb);
          graphicsControl.setTransparentColorFlag(false);
        }
      }
    }
    return true;
  }

  static final int TYPE_TRAILER = 0x3b;
  static final int TYPE_IMAGE_DESCRIPTOR = 0x2c;
  static final int TYPE_EXTENSION_INTRODUCER = 0x21;
  static final int GRAPHIC_CONTROL_LABEL = 0xf9;
  static final int COMMENT_LABEL = 0xfe;
  static final int PLAIN_TEXT_LABEL = 0x01;
  static final int APPLICATION_EXTENSION_LABEL = 0x01;

  class GraphicsControl {
    int pos;

    int delay, transparentColorIndex, disposalMethod;
    boolean userInput, transparentColorFlag;
    int reserved;

    public void setTransparentColorFlag(boolean b) {
      transparentColorFlag = b;
      patchFlags();
    }

    private void patchFlags() {
      patch(pos+1, (reserved << 5)
          | ((disposalMethod & 7) << 2)
          | (userInput ? 2 : 0)
          | (transparentColorFlag ? 1 : 0));
    }
  }

  static class Data {

  };

  static class TableBasedImage extends Data {
    GraphicsControl graphicsControl;
    int left,top,width,height;
    boolean hasLocalColorTable, interlace, sort;
    int localColorTableSize;  // 2 exponent
    public int localColorTablePos; // 0 if none
  }

  static Data OTHER = new Data();

  private Data parseData(int type) throws IOException {
    /*
    <Data> ::=                <Graphic Block>  |
                              <Special-Purpose Block>
  <Graphic Block> ::=       [Graphic Control Extension] <Graphic-Rendering Block>

  <Graphic-Rendering Block> ::=  <Table-Based Image>  |
                                 Plain Text Extension

  <Table-Based Image> ::=   Image Descriptor [Local Color Table] Image Data

  <Special-Purpose Block> ::=    Application Extension  |
                                 Comment Extension


    */

    GraphicsControl gc = null;

    while(type == TYPE_EXTENSION_INTRODUCER) {
      int label = read();

      if(label == GRAPHIC_CONTROL_LABEL) {
        if(gc != null) {
          fail("duplicate grahic control block");
          return null;
        }
        gc = new GraphicsControl();
        gc.pos = currentPos();
        int blockSize = read();
        if(blockSize != 4) {
          fail("unexpected graphic control data block size: "+blockSize);
          return null;
        }
        int flags = read();
        gc.delay = readUnsigned();
        gc.transparentColorIndex = read();
        if(gc.transparentColorIndex == -1) {
          fail("eof in/after graphics control extension");
          return null;
        }

        gc.reserved = (flags >> 5) & 7;
        gc.disposalMethod = (flags >> 2) & 7;
        gc.userInput = (flags & 2) != 0;
        gc.transparentColorFlag = (flags & 1) != 0;

        int terminator = read();
        if(terminator != 0) {
          fail("only one data block expected in graphic control, found size: "+terminator);
          return null;
        }

        type = gifInput.read();
      } else if(label == COMMENT_LABEL) {
        return skipDataBlocks("comment") ? OTHER : null;
      } else if(label == APPLICATION_EXTENSION_LABEL) {
        int blockSize = read();
        // 8 bytes app id
        // 3 bytes mac
        long skipped = gifInput.skip(blockSize);
        if(skipped != blockSize) {
          fail("eof in application extension");
          return null;
        }
        return skipDataBlocks("application extension data") ? OTHER : null;
      } else if(label == PLAIN_TEXT_LABEL) {
        return parsePlainText();
      } else {
        fail("unexpected extension label: "+label);
        return null;
      }

    }
    if(type == TYPE_EXTENSION_INTRODUCER) {
      int label = read();

      if(label == PLAIN_TEXT_LABEL) {
        return parsePlainText();
      } else {
        fail("unexpected extension label: "+label);
        return null;
      }
    } else if (type == TYPE_IMAGE_DESCRIPTOR) {
      TableBasedImage id = new TableBasedImage();
      id.graphicsControl = gc;
      id.left = readUnsigned();
      id.top = readUnsigned();
      id.width = readUnsigned();
      id.height = readUnsigned();
      int flags = read();
      boolean hasLocalColorTable = (flags & (1<<7)) != 0;
      id.interlace = (flags & (1<<6)) != 0;
      id.sort = (flags & (1<<5)) != 0;
      id.localColorTableSize = (flags & 7) + 1;  // 2 exponent

      if(hasLocalColorTable) {
        id.localColorTablePos = currentPos();
        int toSkip = 3 * (1 << id.localColorTableSize);
        long skipped = gifInput.skip(toSkip);
        if(skipped != toSkip) {
          fail("eof in local color table");
          return null;
        }
      }

      return parseImageData() ? id : null;

    }
    fail("unknown block type: "+type);
    return null;
  }

  private Data parsePlainText() throws IOException {
    int blockSize = read();
    long skipped = gifInput.skip(blockSize);
    if(skipped != blockSize) {
      fail("eof in plain text");
      return null;
    }
    return skipDataBlocks("plain text data") ? OTHER : null;
  }

  private boolean parseImageData() throws IOException {
    int minimumCodeSize = read();
    if(minimumCodeSize == -1) {
      fail("eof in image data");
      return false;
    }
    return skipDataBlocks("image data blocks");
  }

  private boolean skipDataBlocks(String context) throws IOException {
    int b;
    while ((b = gifInput.read()) > 0) {
      int toSkip = b;
      long skipped = gifInput.skip(toSkip);
      if(skipped != toSkip) {
        fail("eof in "+context);
        return false;
      }
    }
    if(b == -1) {
      fail("eof in "+context);
      return false;
    }
    return true;
  }
}
